# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.

"""Utilities for wrapping Ansible."""

from __future__ import print_function

import time
import sys

from getpass import getpass
import pexpect
import yaml

from rho import utilities
from rho.utilities import log, ANSIBLE_LOG_PATH, str_to_ascii


def auth_as_ansible_host_vars(auth):
    """Get the Ansible host vars that implement an auth.

    :param auth: the auth. A dictionary with fields 'id', 'name',
        'username', 'password', and 'ssh_key_file'.

    :returns: a dict that can be used as the host variables in an
        Ansible inventory.
    """

    username = auth.get('username')
    password = auth.get('password')
    ssh_key_file = auth.get('ssh_key_file')
    sudo_password = auth.get('sudo_password')

    ansible_vars = {}

    ansible_vars['ansible_user'] = str_to_ascii(username)
    if password:
        ansible_vars['ansible_ssh_pass'] = str_to_ascii(password)
    if ssh_key_file:
        ansible_vars['ansible_ssh_private_key_file'] = \
            str_to_ascii(ssh_key_file)
    if sudo_password:
        ansible_vars['ansible_become_pass'] = sudo_password

    return ansible_vars


def redact_dict(redact_key_list, a_dict):
    """Redact_values in a dictionary

    :param redact_key_list: keys in dictionary that value should be redacted
    :param a_dict: A dictionary to redact
    :returns: the redacted dictionary
    """
    for key in redact_key_list:
        if a_dict is not None and key in a_dict:
            a_dict[key] = utilities.PASSWORD_MASKING
    return a_dict


def log_yaml_inventory(label, inventory):
    """Log yaml inventory but mask passwords

    :param inventory: A dictionary of the ansible inventory
    """

    redact_key_list = ['ansible_become_pass', 'ansible_ssh_pass']

    for group in inventory.keys():
        hosts_dict = inventory[group].get('hosts')
        vars_dict = inventory[group].get('vars')

        for host_dict in hosts_dict.values():
            host_dict = redact_dict(redact_key_list, host_dict)

        vars_dict = redact_dict(redact_key_list, vars_dict)

    log.debug('%s:\n%s', label, yaml.dump(inventory))
    return inventory


class AnsibleProcessException(Exception):
    """Exception raised when an Ansible process fails."""

    pass


class AnsibleTimeoutException(Exception):
    """Exception raised when an Ansible process times out."""

    pass


# pylint: disable=too-many-arguments,too-many-branches
# pylint: disable=too-many-statements,too-many-locals
def run_with_vault(cmd_string, vault_pass, env=None, log_path=None,
                   log_to_stdout=None, log_to_stdout_env=None,
                   ansible_verbosity=0, timeout=None,
                   print_before_run=False, error_on_failure=True):
    """Runs ansible command allowing for password to be provided after
    process triggered.

    Returns after the process completes.

    :param cmd_string: the command to run.
    :param vault_pass: the password to the user's Ansible Vault.
    :param env: the environment to run the subprocess in.
    :param log_path: a path to write the process's log to. Defaults to
        'XDG_DATA_HOME/rho/ansible_log'.
    :param log_to_stdout: if not None, write Ansible's log to stdout using
        the provided function as a filter. Defaults to None.
    :param log_to_stdout_env: a dictionary of environment variables
    :param ansible_verbosity: the number of v's of Ansible verbosity.
    :param timeout: timeout, in seconds, for the called process.

    :param print_before_run: if true, print the command string before running
        it. Defaults to False.
    :returns: the popen.spawn object for the process.
    """

    # pexpect provides the ability to send the process's output to a
    # single Python file object. We want to send it to a file and
    # maybe also stdout. The solution is to have pexpect log to the
    # file and then use 'tail -f' to copy that to stdout.
    tail_process = None
    if not log_path:
        log_path = ANSIBLE_LOG_PATH

    if ansible_verbosity:
        cmd_string = cmd_string + ' -' + 'v' * ansible_verbosity

    try:
        utilities.ensure_data_dir_exists()
        with open(log_path, 'wb') as logfile:
            pass
        with open(log_path, 'r+b') as logfile:
            log.debug('Running Ansible: %s', cmd_string)
            if print_before_run:
                print('Running:', cmd_string)
            child = pexpect.spawn(cmd_string, timeout=timeout,
                                  env=env)

            if log_to_stdout is not None:
                try:
                    tail_process = utilities.tail_log(log_path,
                                                      ansible_verbosity,
                                                      log_to_stdout,
                                                      log_to_stdout_env)
                except ValueError as ex:
                    log.error('Unable to tail Ansible log: %s', ex)

            child.expect('Vault password:')
            child.sendline(vault_pass)
            # Set the log file *after* we send the user's Vault
            # password to Ansible, so we don't log the password.
            child.logfile = logfile
            last_pos = logfile.tell()

            i = child.expect([pexpect.EOF, 'Enter passphrase for key .*:',
                              'you want to continue connecting (yes/no)?'])
            while i:
                new_pos = logfile.tell()
                logfile.seek(last_pos)
                logfile_lines = logfile.readlines()
                log.info(logfile_lines)
                print(logfile_lines[-1].replace('\r\n', ''))
                logfile.seek(new_pos)
                last_pos = new_pos
                child.logfile = None
                # Ansible has already printed a prompt; it would be
                # confusing if getpass printed another one.
                child.sendline(getpass(''))
                child.logfile = logfile
                i = child.expect([pexpect.EOF, 'Enter passphrase for key .*:',
                                  'you want to continue connecting (yes/no)?'])

            if child.isalive():
                child.wait()
            if log_to_stdout:
                time.sleep(2)

    except pexpect.EOF:
        print('Error: unexpected Ansible output')
        if tail_process is not None:
            tail_process.terminate()
        sys.exit(1)
    except pexpect.TIMEOUT:
        child.terminate()
        if tail_process is not None:
            tail_process.terminate()

        if not timeout:
            print('Error: unexpected Ansible output')
            sys.exit(1)

        raise AnsibleTimeoutException()

    if tail_process is not None:
        tail_process.terminate()
    if (child.exitstatus != 0 and child.exitstatus != 4) or child.signalstatus:
        if error_on_failure is False and child.exitstatus == 2:
            pass
        else:
            raise AnsibleProcessException(
                'Ansible process failed with status %s, signal status %s' %
                (child.exitstatus, child.signalstatus))
